/*
 * binder interface for wpa_supplicant daemon
 * Copyright (c) 2004-2016, Jouni Malinen <j@w1.fi>
 * Copyright (c) 2004-2016, Roshan Pius <rpius@google.com>
 *
 * This software may be distributed under the terms of the BSD license.
 * See README for more details.
 */

#ifndef WPA_SUPPLICANT_BINDER_BINDER_CONSTANTS_H
#define WPA_SUPPLICANT_BINDER_BINDER_CONSTANTS_H

namespace wpa_supplicant_binder {
namespace binder_constants {

extern const char kServiceName[];

} /* namespace binder_constants */
} /* namespace wpa_supplicant_binder */

#endif /* WPA_SUPPLICANT_BINDER_BINDER_CONSTANTS_H */
