/*
 * binder interface for wpa_supplicant daemon
 * Copyright (c) 2004-2016, Jouni Malinen <j@w1.fi>
 * Copyright (c) 2004-2016, Roshan Pius <rpius@google.com>
 *
 * This software may be distributed under the terms of the BSD license.
 * See README for more details.
 */

#include "binder_constants.h"

namespace wpa_supplicant_binder {
namespace binder_constants {

const char kServiceName[] = "wpa_supplicant";

} /* namespace binder_constants */
} /* namespace wpa_supplicant_binder */
